// alien_overlay.c
// livido plugin
// (c) G. Finch (salsaman) 2005
//
// released under the GNU GPL 2
// see file COPYING or www.gnu.org for details








#include "plugins-livido.h"

livido_setup_f livido_setup;

livido_init_f alien_over_init;
livido_deinit_f alien_over_deinit;
livido_process_f alien_over_process;


typedef struct {
  int inited;
  unsigned char *old_pixel_data;
} static_data;

/* declare our init function */
int alien_over_init(livido_instance_t *inst) {
  int rowstride=inst->in_channels->rowstrides[0];
  int size=inst->in_channels->height*rowstride;
  static_data *sdata=(static_data *)(*inst->malloc) (sizeof(static_data));

  sdata->old_pixel_data=(unsigned char *)(*inst->malloc)(size);
  sdata->inited=0;
  inst->internal=sdata;
  return LIVIDO_NO_ERROR;
}


int alien_over_process
(livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  // host should have returned one input channel and one output channel
  // (which can be NULL, since we exported the property INPLACE)

  // we also set our input/output channels same as, so we can assume both frame
  // sizes and palettes match

  int i,j,k;
  int width=inst->in_channels->width*3;
  int rowstride_i=inst->in_channels->rowstrides[0];
  int size=inst->in_channels->height*rowstride_i;

  unsigned char *curr_pixel_data, *new_pixel_data, byte;
  int rowstride_o;

  int inplace=(out==NULL);

  static_data *sdata=(static_data *)(inst->internal);
  unsigned char *old_pixel_data=sdata->old_pixel_data;

  curr_pixel_data=in->pixel_data[0];

  if (inplace) {
    new_pixel_data=curr_pixel_data;
    rowstride_o=rowstride_i;
  }
  else {
    new_pixel_data=out->pixel_data[0];
    rowstride_o=inst->out_channels->rowstrides[0];
  }


  for (i=0;i<size;i+=rowstride_i) {
    for (j=0;j<width;j++) {
      k=i/rowstride_i*rowstride_o;
      byte=curr_pixel_data[i+j];
      if (sdata->inited) {
	new_pixel_data[k+j]=((char)(byte=curr_pixel_data[i+j])+(char)old_pixel_data[i+j])>>1;
      }
      else if (!inplace) new_pixel_data[k+j]=curr_pixel_data[i+j];
      old_pixel_data[i+j]=byte;
    }
  }
  sdata->inited=1;
  return LIVIDO_NO_ERROR;
}


int alien_over_deinit (livido_instance_t *inst) {
  static_data *sdata;
  if (inst==NULL) return LIVIDO_ERROR_NULL_INSTANCE;
  sdata=(static_data *)(inst->internal);
  if (sdata!=NULL) {
    (*inst->free)(sdata->old_pixel_data);
    (*inst->free) (sdata);
    inst->internal=NULL;
  }
  return LIVIDO_NO_ERROR; // success
}


// create 1 input and one matching output channel, no masks

// allow RGB888 and BGR888 palettes
livido_palette_t in_palettes[]={LIVIDO_PALETTE_RGB888,LIVIDO_PALETTE_BGR888,LIVIDO_PALETTE_END};
livido_palette_t out_palettes[]={LIVIDO_PALETTE_RGB888,LIVIDO_PALETTE_BGR888,LIVIDO_PALETTE_END};

livido_channel_template_t in_channels={"In 1",LIVIDO_CHANNEL_HOST_CAN_CHANGE_PALETTE,in_palettes,0,0,0,NULL};
livido_channel_template_t out_channels={"Out 1",LIVIDO_CHANNEL_SAME_AS_SIZE|LIVIDO_CHANNEL_SAME_AS_PALETTE|LIVIDO_CHANNEL_HOST_CAN_CHANGE_PALETTE,out_palettes,1,0,0,NULL};

livido_instance_template_t fx={"alien_overlay","salsaman","Adds the current frame to the previous frame",1,1.7,LIVIDO_PROPERTY_REALTIME|LIVIDO_PROPERTY_CAN_DO_INPLACE,NULL,NULL,&in_channels,&out_channels,alien_over_init,alien_over_deinit,alien_over_process,NULL,NULL,NULL};


livido_instance_template_t *livido_setup(void) {
  return &fx;
}

