// mirrors.c
// livido plugin
// (c) G. Finch (salsaman) 2005
//
// released under the GNU GPL 2
// see file COPYING or www.gnu.org for details

#include "plugins-livido.h"

livido_setup_f livido_setup;

livido_init_f mirror_init;
livido_deinit_f mirror_deinit;
livido_process_f mirrorx_process;
livido_process_f mirrory_process;
livido_process_f mirrorxy_process;



/* declare our init function */
int mirror_init(livido_instance_t *inst) {
  return LIVIDO_NO_ERROR;
}


int mirrorx_process
(livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  // host should have returned one input channel and one output channel
  // (which can be NULL, since we exported the property INPLACE)

  // we also set our input/output channels same as, so we can assume both frame
  // sizes and palettes match

  int i,j,offset=0;
  int rowstride=inst->in_channels->rowstrides[0];
  int width=inst->in_channels->width*3;
  int size=inst->in_channels->height*rowstride;

  unsigned char *old_pixel_data=in->pixel_data[0], *new_pixel_data;
  livido_palette_t palette=inst->in_channels->palette;

  int inplace=(out==NULL);

  if (palette==LIVIDO_PALETTE_BGR888) offset=-1;

  if (inplace) {
    new_pixel_data=old_pixel_data;
  }
  else new_pixel_data=out->pixel_data[0];


  for (i=0;i<size;i+=rowstride) {
    for (j=0;j<width/2;j++) {
      new_pixel_data[i+width-1-j+offset*2]=old_pixel_data[i+j];
      if (!inplace) new_pixel_data[i+j]=old_pixel_data[i+j];
      if (palette==LIVIDO_PALETTE_BGR888) offset=++offset<2?offset:-1;
    }
  } 
  return LIVIDO_NO_ERROR;
}



int mirrory_process
(livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  // host should have returned one input channel and one output channel
  // (which can be NULL, since we exported the property INPLACE)

  // we also set our input/output channels same as, so we can assume both frame
  // sizes and palettes match

  int i,j;
  int rowstride=inst->in_channels->rowstrides[0];
  int width=inst->in_channels->width*3;
  int size=inst->in_channels->height*rowstride;

  unsigned char *old_pixel_data=in->pixel_data[0], *new_pixel_data;

  int inplace=(out==NULL);

  if (inplace) {
    new_pixel_data=old_pixel_data;
  }
  else new_pixel_data=out->pixel_data[0];

  for (i=0;i<size/2;i+=rowstride) {
    for (j=0;j<width;j++) {
      new_pixel_data[size-rowstride-i+j]=old_pixel_data[i+j];
      if (!inplace) new_pixel_data[i+j]=old_pixel_data[i+j];
    }
  } 
  return LIVIDO_NO_ERROR;
}



int mirrorxy_process
(livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  int ret=mirrorx_process(inst,in,out);
  if (ret) return ret;
  return mirrory_process(inst,in,out);
}






int mirror_deinit (livido_instance_t *inst) {
  return LIVIDO_NO_ERROR; // success
}


// create 1 input and one matching output channel, no masks

// allow RGB888 and BGR888 palettes

livido_palette_t in_palettes[]={LIVIDO_PALETTE_RGB888,LIVIDO_PALETTE_BGR888,LIVIDO_PALETTE_END};
livido_palette_t out_palettes[]={LIVIDO_PALETTE_RGB888,LIVIDO_PALETTE_BGR888,LIVIDO_PALETTE_END};

livido_channel_template_t in_channels={"In 1",LIVIDO_CHANNEL_HOST_CAN_RESIZE,in_palettes,0,0,0,NULL};
livido_channel_template_t out_channels={"Out 1",LIVIDO_CHANNEL_SAME_AS_SIZE|LIVIDO_CHANNEL_SAME_AS_PALETTE|LIVIDO_CHANNEL_HOST_CAN_RESIZE,out_palettes,1,0,0,NULL};

livido_instance_template_t fx[]={{"mirrorx","salsaman","Mirror left to right",1,1.7,LIVIDO_PROPERTY_REALTIME|LIVIDO_PROPERTY_CAN_DO_INPLACE,NULL,NULL,&in_channels,&out_channels,mirror_init,mirror_deinit,mirrorx_process,NULL,NULL,NULL},{"mirrory","salsaman","Mirror top to bottom",1,1.7,LIVIDO_PROPERTY_REALTIME|LIVIDO_PROPERTY_CAN_DO_INPLACE,NULL,NULL,&in_channels,&out_channels,mirror_init,mirror_deinit,mirrory_process,NULL,NULL,NULL},{"mirrorxy","salsaman","Mirror top to bottom and left to right",1,1.7,LIVIDO_PROPERTY_REALTIME|LIVIDO_PROPERTY_CAN_DO_INPLACE,NULL,NULL,&in_channels,&out_channels,mirror_init,mirror_deinit,mirrorxy_process,NULL,NULL,NULL}};


livido_instance_template_t *livido_setup(void) {
  fx[0].next=&fx[1];
  fx[1].next=&fx[2];
  return fx;
}

