/*
 * EffecTV - Realtime Digital Video Effector
 * Copyright (C) 2001-2005 FUKUCHI Kentaro
 *
 * RippleTV - Water ripple effect.
 * Copyright (C) 2001-2002 FUKUCHI Kentaro
 *
 */

/* modified for Livido by G. Finch (salsaman)
   modifications (c) G. Finch */



//#define DEBUG

#define MAGIC_THRESHOLD 70

#include "plugins-livido.h"
#include "livido-lib.c"

livido_setup_f livido_setup;

livido_init_f ripple_init;
livido_deinit_f ripple_deinit;
livido_process_f ripple_process;

typedef unsigned int RGB32;


// TODO - handle rowstrides

// TODO

static int sqrtable[256];

static const int point = 16;
static const int impact = 2;
static const int decay = 8;
static const int loopnum = 2;
static int period = 0;
static int rain_stat = 0;
static unsigned int drop_prob = 0;
static int drop_prob_increment = 0;
static int drops_per_frame_max = 0;
static int drops_per_frame = 0;
static int drop_power = 0;

/////////////



struct _sdata {
  int *map;
  int *map1;
  int *map2;
  int *map3;
  int bgIsSet;
  signed char *vtable;
  short *background;
  unsigned char *diff;
  int threshold;
  uint32_t fastrand_val;
};








static void image_bgset_y(RGB32 *src, int width, int height, struct _sdata *sdata)
{
  register int i;
  int R, G, B;
  RGB32 *p;
  short *q;
  int video_area=width*height;

  p = src;
  q = sdata->background;
  for(i=0; i<video_area; i++) {
    R = ((*p)&0xff0000)>>(16-1);
    G = ((*p)&0xff00)>>(8-2);
    B = (*p)&0xff;
    *q = (short)(R + G + B);
    p++;
    q++;
  }
}


/* Background image is refreshed every frame */
static void image_bgsubtract_update_y(RGB32 *src, int width, int height, struct _sdata *sdata)
{
  register int i;
  int R, G, B;
  RGB32 *p;
  short *q;
  unsigned char *r;
  int v;
  int video_area=width*height;
  
  p = src;
  q = sdata->background;
  r = sdata->diff;
  for(i=0; i<video_area; i++) {
    R = ((*p)&0xff0000)>>(16-1);
    G = ((*p)&0xff00)>>(8-2);
    B = (*p)&0xff;
    v = (R + G + B) - (int)(*q);
    *q = (short)(R + G + B);
    *r = ((v + sdata->threshold)>>24) | ((sdata->threshold - v)>>24);
    
    p++;
    q++;
    r++;
  }
}


/////////////////////////////////////////////////////

inline uint32_t fastrand(struct _sdata *sdata)
{
#define rand_a 1073741789L
#define rand_c 32749L

  return (sdata->fastrand_val = rand_a * sdata->fastrand_val + rand_c);
}


static void setTable()
{
	int i;

	for(i=0; i<128; i++) {
		sqrtable[i] = i*i;
	}
	for(i=1; i<=128; i++) {
		sqrtable[256-i] = -i*i;
	}
}

static int setBackground(RGB32 *src, int width, int height, struct _sdata *sdata)
{
	image_bgset_y(src, width, height, sdata);
	sdata->bgIsSet = 1;

	return 0;
}


///////////////////////////////////////////////////////




livido_palette_t in_palettes[]={LIVIDO_PALETTE_RGB32,LIVIDO_PALETTE_END};
livido_palette_t out_palettes[]={LIVIDO_PALETTE_RGB32,LIVIDO_PALETTE_END};

livido_channel_template_t in_channels={"In 1",0,in_palettes,0,0,0,NULL};
livido_channel_template_t out_channels={"Out 1",LIVIDO_CHANNEL_SAME_AS_SIZE|LIVIDO_CHANNEL_SAME_AS_PALETTE,out_palettes,1,0,0,NULL};

livido_instance_template_t fx={"rippleTV","effectTV","Add ripples and rain to video frames",1,1.7,LIVIDO_PROPERTY_REALTIME,NULL,NULL,&in_channels,&out_channels,ripple_init,ripple_deinit,ripple_process,NULL,NULL,NULL};




livido_instance_template_t *livido_setup(void) {
  setTable();
  return &fx;
}



int ripple_init(livido_instance_t *inst) {
  int map_h;
  int map_w;
  struct _sdata *sdata;
  char *list[2]={"ripples","rain"};

  if (inst==NULL) return LIVIDO_ERROR_NULL_INSTANCE;

  map_h=inst->in_channels[0].height;
  map_w=inst->in_channels[0].width;

  sdata=malloc(sizeof(struct _sdata));

  if(sdata == NULL ) {
    return LIVIDO_ERROR_MEMORY_ALLOCATION;
  }

  sdata->map = (int *)malloc(map_h*map_w*3*sizeof(int));
  if(sdata->map == NULL ) {
    free(sdata);
    return LIVIDO_ERROR_MEMORY_ALLOCATION;
  }
  sdata->vtable = (signed char *)malloc(map_h*map_w*2*sizeof(signed char));
  if(sdata->vtable == NULL ) {
    free(sdata->map);
    free(sdata);
    return LIVIDO_ERROR_MEMORY_ALLOCATION;
  }
  sdata->background = (short *)malloc(map_h*map_w*sizeof(short));
  if(sdata->background == NULL ) {
    free(sdata->vtable);
    free(sdata->map);
    free(sdata);
    return LIVIDO_ERROR_MEMORY_ALLOCATION;
  }
  sdata->diff = (unsigned char *)malloc(map_h*map_w * 4 * sizeof(unsigned char));
  if(sdata->diff == NULL ) {
    free(sdata->background);
    free(sdata->vtable);
    free(sdata->map);
    free(sdata);
    return LIVIDO_ERROR_MEMORY_ALLOCATION;
  }

  memset(sdata->map, 0, map_h*map_w*3*sizeof(int));
  memset(sdata->vtable, 0, map_h*map_w*2*sizeof(signed char));
  memset(sdata->diff, 0, map_h*map_w*4*sizeof(unsigned char));
  sdata->map1 = sdata->map;
  sdata->map2 = sdata->map + map_h*map_w;
  sdata->map3 = sdata->map + map_w * map_h * 2;
  sdata->bgIsSet = 0;
  sdata->threshold=MAGIC_THRESHOLD*7;
  sdata->fastrand_val=0;

  inst->internal=sdata;

  inst->num_in_parameters=1;
  inst->in_parameters=malloc(sizeof(livido_parameter_t));
  inst->in_parameters[0]=livido_init_string_list_param("Type",0,2,list);

  return LIVIDO_NO_ERROR;
}

int ripple_deinit (livido_instance_t *inst) {
  struct _sdata *sdata;
  if (inst==NULL) return LIVIDO_ERROR_NULL_INSTANCE;
  sdata=inst->internal;
  if (sdata != NULL) {
    free(sdata->diff);
    free(sdata->background);
    free(sdata->vtable);
    free(sdata->map);
    free(sdata);
  }
  return LIVIDO_NO_ERROR;
}


static void motiondetect(RGB32 *src, int width, int height, struct _sdata *sdata)
{
  unsigned char *diff;
  int *p, *q;
  register int x, y;
  int h;
  
  if(!sdata->bgIsSet) {
    setBackground(src,width,height,sdata);
  }
  image_bgsubtract_update_y(src,width,height,sdata);

  diff = sdata->diff;
  p = sdata->map1+width+1;
  q = sdata->map2+width+1;
  diff += width+2;

  for(y=height-2; y>0; y--) {
    for(x=width-2; x>0; x--) {
      h = (int)*diff + (int)*(diff+1) + (int)*(diff+width) + (int)*(diff+width+1);
      if(h>0) {
	*p = h<<(point + impact - 8);
	*q = *p;
      }
      p++;
      q++;
      diff += 2;
    }
    diff += width+2;
    p+=2;
    q+=2;
  }
}



static inline void drop(int power, int width, int height, struct _sdata *sdata)
{
  int x, y;
  int *p, *q;

  x = fastrand(sdata)%(width-4)+2;
  y = fastrand(sdata)%(height-4)+2;
  p = sdata->map1 + y*width + x;
  q = sdata->map2 + y*width + x;
  *p = power;
  *q = power;
  *(p-width) = *(p-1) = *(p+1) = *(p+width) = power/2;
  *(p-width-1) = *(p-width+1) = *(p+width-1) = *(p+width+1) = power/4;
  *(q-width) = *(q-1) = *(q+1) = *(q+width) = power/2;
  *(q-width-1) = *(q-width+1) = *(q+width-1) = *(p+width+1) = power/4;
}

static void raindrop(int width, int height, struct _sdata *sdata)
{
  
  int i;
  
  if(period == 0) {
    switch(rain_stat) {
    case 0:
      period = (fastrand(sdata)>>23)+100;
      drop_prob = 0;
      drop_prob_increment = 0x00ffffff/period;
      drop_power = (-(fastrand(sdata)>>28)-2)<<point;
      drops_per_frame_max = 2<<(fastrand(sdata)>>30); // 2,4,8 or 16
      rain_stat = 1;
      break;
    case 1:
      drop_prob = 0x00ffffff;
      drops_per_frame = 1;
      drop_prob_increment = 1;
      period = (drops_per_frame_max - 1) * 16;
      rain_stat = 2;
      break;
    case 2:
      period = (fastrand(sdata)>>22)+1000;
      drop_prob_increment = 0;
      rain_stat = 3;
      break;
    case 3:
      period = (drops_per_frame_max - 1) * 16;
      drop_prob_increment = -1;
      rain_stat = 4;
      break;
    case 4:
      period = (fastrand(sdata)>>24)+60;
      drop_prob_increment = -(drop_prob/period);
      rain_stat = 5;
      break;
    case 5:
    default:
      period = (fastrand(sdata)>>23)+500;
      drop_prob = 0;
      rain_stat = 0;
      break;
    }
  }
  switch(rain_stat) {
  default:
  case 0:
    break;
  case 1:
  case 5:
    if((fastrand(sdata)>>8)<drop_prob) {
      drop(drop_power,width,height,sdata);
    }
    drop_prob += drop_prob_increment;
    break;
  case 2:
  case 3:
  case 4:
    for(i=drops_per_frame/16; i>0; i--) {
      drop(drop_power,width,height,sdata);
    }
    drops_per_frame += drop_prob_increment;
    break;
  }
  period--;
}




int ripple_process (livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  register int x, y, i;
  int dx, dy;
  int h, v;
  int width, height;
  int *p, *q, *r;
  signed char *vp;
  struct _sdata *sdata;
  RGB32 *src,*dest;
  int mode;

  if (inst==NULL) return LIVIDO_ERROR_NULL_INSTANCE;

  livido_get_int32_prop(inst->in_parameters[0],"value",&mode);

  width = inst->in_channels[0].width;
  height = inst->in_channels[0].height;

  sdata=inst->internal;
  src=in->pixel_data[0];
  dest=out->pixel_data[0];

  /* impact from the motion or rain drop */
  if(mode) {
    raindrop(width,height,sdata);
  } else {
    motiondetect(src,width,height,sdata);
  }
  
  /* simulate surface wave */
	
  /* This function is called only 30 times per second. To increase a speed
   * of wave, iterates this loop several times. */
  for(i=loopnum; i>0; i--) {
    /* wave simulation */
    p = sdata->map1 + width + 1;
    q = sdata->map2 + width + 1;
    r = sdata->map3 + width + 1;
    for(y=height-2; y>0; y--) {
      for(x=width-2; x>0; x--) {
	h = *(p-width-1) + *(p-width+1) + *(p+width-1) + *(p+width+1)
	  + *(p-width) + *(p-1) + *(p+1) + *(p+width) - (*p)*9;
	h = h >> 3;
	v = *p - *q;
	v += h - (v >> decay);
	*r = v + *p;
	p++;
	q++;
	r++;
      }
      p += 2;
      q += 2;
      r += 2;
    }
    
    /* low pass filter */
    p = sdata->map3 + width + 1;
    q = sdata->map2 + width + 1;
    for(y=height-2; y>0; y--) {
      for(x=width-2; x>0; x--) {
	h = *(p-width) + *(p-1) + *(p+1) + *(p+width) + (*p)*60;
	*q = h >> 6;
	p++;
	q++;
      }
      p+=2;
      q+=2;
    }

    p = sdata->map1;
    sdata->map1 = sdata->map2;
    sdata->map2 = p;
  }

  vp = sdata->vtable;
  p = sdata->map1;
  for(y=height-1; y>0; y--) {
    for(x=width-1; x>0; x--) {
      /* difference of the height between two voxel. They are twiced to
       * emphasise the wave. */
      vp[0] = sqrtable[((p[0] - p[1])>>(point-1))&0xff]; 
      vp[1] = sqrtable[((p[0] - p[width])>>(point-1))&0xff]; 
      p++;
      vp+=2;
    }
    p++;
    vp+=2;
  }
  
  vp = sdata->vtable;

  /* draw refracted image. The vector table is stretched. */
  for(y=0; y<height; y+=2) {
    for(x=0; x<width; x+=2) {
      h = (int)vp[0];
      v = (int)vp[1];
      dx = x + h;
      dy = y + v;
      if(dx<0) dx=0;
      if(dy<0) dy=0;
      if(dx>=width) dx=width-1;
      if(dy>=height) dy=height-1;
      dest[0] = src[dy*width+dx];
      
      i = dx;
      
      dx = x + 1 + (h+(int)vp[2])/2;
      if(dx<0) dx=0;
      if(dx>=width) dx=width-1;
      dest[1] = src[dy*width+dx];
      
      dy = y + 1 + (v+(int)vp[width*2+1])/2;
      if(dy<0) dy=0;
      if(dy>=height) dy=height-1;
      dest[width] = src[dy*width+i];

      dest[width+1] = src[dy*width+dx];
      dest+=2;
      vp+=2;
    }
    dest += width;
    vp += 2;
  }
  return LIVIDO_NO_ERROR;
}

